/*
 Main routines for eliberation, a tool to rescue books from the REB120 Ebook
 This code is released into the public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef _MSC_VER
#include <windows.h>
#else
#include <dirent.h>
#endif

#include "aes.h"
#include "defs.h"
#include "fat12.h"

#ifndef _MSC_VER
#define strcmpi	 strcasecmp
#endif

#ifndef NAME_MAX
#define NAME_MAX 256
#endif

extern int read_pref_string(byte * prefp, byte * name, byte * buf, int buflen);
int  global_errors;
byte * global_prefs = NULL;
int  has_aes_key = 0;
byte global_aes_key[16];

char * global_prefs_file = "nuflash.img";

byte * read_prefs_from_fat_image( char * fat_image) {
    
    fatso   fat;
    fatdir  *dir, *pdir;
    fatentry fe;
    byte    * retp, * tmp;
    int     index;

    dir = pdir = NULL;
    tmp = retp =  NULL;
    if (fat_birth(&fat, fat_image)) return NULL;
    
    dir = fat_directory(&fat,NULL);
    if (!dir) goto l;

    index = 0;
    do {
        if (fat_entry(dir,index,&fe)) goto l;
        index++;
    } while (strcmp(fe.name,"PREFS   RES") != 0);
   
    pdir = fat_directory(&fat,&fe);
    free(dir); dir = NULL;
    index = 0;
    while (1) {
        if (fat_entry(pdir,index,&fe)) goto l;
        index++;
        /* Filter out long file names, directories and junk */
        if ((fe.attributes & 0x18) || (fe.cluster == 0) || (fe.size < 14)) 
            continue;
        tmp = fat_file(&fat,&fe);
        if (!tmp) goto l;
        if (memcmp(tmp+sizeof(dword),"\0\1PREF",6) == 0) break;
        free(tmp);
    } 
    if (!tmp) goto l;

    retp = tmp; tmp = NULL;
l:
    if (tmp) free(tmp);
    if (dir) free(dir);
    if (pdir) free(pdir);
    fat_death(&fat);
    return retp;
}

int get_book_key_from_prefs(char * s,byte * book_key) 
{
    byte aes_buffer[16];
    int r, i;
    struct aes_ctx ctx;

    if (!has_aes_key) { 
       fprintf(stderr, "No AES key present. Verify your key file.\n");  
       return -1;
    }
    if (!global_prefs) {

       if ( (!global_prefs_file)  || 
       (NULL == (global_prefs=read_prefs_from_fat_image(global_prefs_file)))) {
          fprintf(stderr,"No key in book, and can't load the preferences.\n");
          return -1;
       }
    }
    r = read_pref_string(global_prefs, s, aes_buffer,sizeof(aes_buffer));
    if (r) return -1;
    aes_set_key(&ctx, global_aes_key, sizeof(global_aes_key));
    aes_decrypt(&ctx,&aes_buffer[0],&aes_buffer[0]);
    hexdump(aes_buffer, 16);
    for (i = 8; i < sizeof(aes_buffer); i++) {
        if (aes_buffer[i] != '\0') {
            return -1;
        }
    }
    memcpy(book_key, aes_buffer, 8);
    return 0; 
}

#ifdef _MSC_VER
int foreach_file(char * path_param, int (*callback)(char *,void *), void * data){
    WIN32_FIND_DATA FileData;
    HANDLE hList;
    int len;
    char * pathname;
    char pathbuffer[PATHNAME_MAX + NAME_MAX + 2];
    int err;

    len = strlen(path_param);
    if (len > PATHNAME_MAX)
    {
        fprintf(stderr,"Path name is %d, longer than max of %d\n", len,
            PATHNAME_MAX);
        return -1;
    }
    pathname = &pathbuffer[PATHNAME_MAX - len];
    pathbuffer[PATHNAME_MAX] = '*';
    pathbuffer[PATHNAME_MAX+1] = '\0';
    if (path_param[len - 1] != '\\')
    {
        pathname--;
        pathbuffer[PATHNAME_MAX - 1] = '\\';
    }
    strncpy(pathname,path_param, len);

    // Get the first file
    hList = FindFirstFile(pathname, &FileData);
    if (hList == INVALID_HANDLE_VALUE)
    { 
        return -1;
    }
    do { 
        if ((strcmp(FileData.cFileName,".") == 0) || 
            (strcmp(FileData.cFileName,"..") == 0)) continue;
        strncpy(&pathbuffer[PATHNAME_MAX],FileData.cFileName, NAME_MAX-1);
        pathbuffer[PATHNAME_MAX+NAME_MAX] = '\0';
        err = callback(pathname, data);
        if (err < 0) return err;
    } while (FindNextFile(hList, &FileData) != 0);
    FindClose(hList);
    return 0;
}
#else

int foreach_file(char * path_param, int (*callback)(char *,void *), void * data){
    DIR * dir;
    struct dirent * entry;
    int len;
    char * pathname;
    char pathbuffer[PATHNAME_MAX + NAME_MAX];
    int err;

    len = strlen(path_param);
    if (len > PATHNAME_MAX)
    {
        fprintf(stderr,"Path name is %d, longer than max of %d\n", len,
            PATHNAME_MAX);
        return -1;
    }
    pathname = &pathbuffer[PATHNAME_MAX - len];
    pathbuffer[PATHNAME_MAX] = '\0';
    if (path_param[len - 1] != '/')
    {
        pathname--;
        pathbuffer[PATHNAME_MAX - 1] = '/';
    }
    strncpy(pathname,path_param, len);

    dir = opendir(pathname);
    if (!dir)
    {
        perror("opendir()");
        return -1;
    }

    while (entry = readdir(dir))
    {
        strncpy(&pathbuffer[PATHNAME_MAX], entry->d_name, NAME_MAX);
        err = callback(pathname, data);
        if (err < 0) return err;
    }
    return 0;
}
#endif

int resdircallback(char *s , void *v)
{
    int len; 
    int r;

    len = strlen(s);
    if (strcmpi(&s[len - 4],".RES") == 0) {
        r = res2imp(s, v);
        if (r < 0) {  
            global_errors++;
            fprintf(stderr,"FAILED TO CONVERT \"%s\" !\n", s);
        }
    }
    return 0;
}

int read_whole_file(char * filename, unsigned char ** ptr)
{
    FILE    * f;
    int     bytecount, filelen;
    unsigned char * fileptr;

    f = fopen(filename,"rb");
    if (!f)
    {
        if (errno == EISDIR) return 0;
        perror("fopen() ");
        fprintf(stderr, "Error with \"%s\" (%d)\n", filename, errno);
        return -1;
    }
    fseek(f, 0, SEEK_END);
    filelen = ftell(f);
    fseek(f, 0, SEEK_SET);

    if (!filelen) { fclose(f); return 0; }

    /* printf("Opened %s (%d bytes)\n", filename, filelen); */

    fileptr = (unsigned char *)malloc(filelen);
    if (!fileptr)
    {
        fprintf(stderr,"Unable to allocate %d chars for file \"%s\"!\n",
            filelen, filename);
        return -1;
    }

    bytecount = fread(fileptr,1, filelen, f);
    if (bytecount != filelen)
    {
        /* Directories under Linux are wierd... they return a size of
         * 4096, but only read a length of 0 */
        if (!bytecount)
        {
            free(fileptr);
            fclose(f);
            return 0;
        }
        fprintf(stderr,"Read %d out of %d chars for file \"%s\"!\n",
            bytecount, filelen, filename);
        return -1;
    }
    fclose(f);
    if (ptr) *ptr = fileptr;
    return filelen;
}


int parsehex(char * s, char * b, int nbytes)
{
    int len, i;
    unsigned char u;
    unsigned char * p;

    len = strlen(s);
    if ((len/2) > nbytes)  {
        fprintf(stderr,"Too many bytes for simple parser.\n"); exit(-1);
    }
    p = b;
    printf("%s\n", s);
    for (i = 0; i < len; i+= 2)
    {
        u = 0;

        if ( (*s >= 'A') && (*s <= 'F')) { u = *(s++) - ('A' - 10); }
        else if ( (*s >= 'a') && (*s <= 'f')) { u = *(s++) - ('a' - 10); }
        else if ( (*s >= '0') && (*s <= '9')) { u = *(s++) - '0'; }
        else {
            fprintf(stderr,"Bad character '%c' in hex string.\n", *s);
            exit(-1);
        }

        u <<= 4;
        if ( (*s >= 'A') && (*s <= 'F')) { u += *(s++) - ('A' - 10); }
        else if ( (*s >= 'a') && (*s <= 'f')) { u += *(s++) - ('a' - 10); }
        else if ( (*s >= '0') && (*s <= '9')) { u += *(s++) - '0'; }
        else {
            fprintf(stderr,"Bad character '%c' in hex string.\n", *s);
            exit(-1);
        }

        *(p++) = u;
    }
    return len/2;
}

#define MAX_NAKED_ARG 1

int main(int argc, char ** argv) {
    int r, i;
    int need_help = 0;
    char * args[MAX_NAKED_ARG];
    int naked_arg_count = 0;
    char * imgfile = NULL, * keyfile = NULL;
    byte des_key[8], * desp = NULL;

    i = 1;
    while ( i < argc ) {
        if (argv[i][0] != '-') {
            if (naked_arg_count >= MAX_NAKED_ARG) { need_help++; break; }
            args[naked_arg_count++] = argv[i];
            i++;
            continue;
        }
        switch (toupper(argv[i][1])) {
        case 'F': i++; if (i < argc) { imgfile = argv[i]; } break;
        case 'K': i++; if (i < argc) { keyfile = argv[i]; } break;
        case 'H': need_help++; break;
        }
        i++;
    }
    if ((naked_arg_count != 1) || (need_help)) {
        fprintf(stderr,"eliberate Version 1.0\n");
        fprintf(stderr,"Basic usage: eliberate G:\\SOFTBKS \n");
        fprintf(stderr,"Options:\n"); 
        fprintf(stderr,"   -h  		This message\n");
        fprintf(stderr,"   -f filename  Change image filename\n");
        fprintf(stderr,"   -k filename  Change key filename\n");
        exit(-1);
    }

    if (imgfile) {
        global_prefs_file = imgfile;
    }

    if (!keyfile) keyfile = "reb1200.key";

    desp = NULL;
    {
        int keylen;
        byte * keyp = NULL;;
        keylen  = read_whole_file(keyfile, &keyp);
        if (keylen == 8) { /* Old version */
            if (keyp) {
                memcpy(&des_key[0], keyp, 8); 
                desp = &des_key[0];
                free(keyp);
            }
        } else  if (keylen == 32) { /* New version, 2 keys */
            if (keyp) {
                desp = &des_key[0];
                memcpy(&des_key[0], keyp+12, 4);
                memcpy(&des_key[4], keyp+8, 4);
                has_aes_key++;
                memcpy(&global_aes_key[0],keyp+16, 16);
                free(keyp);
            }

        } 
        /* Otherwise, no key, but no worries, yet. */
    }
    global_errors = 0;
    foreach_file(args[0], resdircallback, desp);
    if (global_prefs) free(global_prefs);
    exit(global_errors);
} 
